"""
Projector command definitions

To update this file with the full list of command classes automatically, run:
`python tools/update_imports.py`
"""

# ruff: noqa: F401

from .command import (
    Anamorphic,
    Aspect,
    ClearMotionDrive,
    ColorDepth,
    Colorimetry,
    ColorManagement,
    ColorProfile,
    ColorSpace,
    ColorTemperature,
    ColorTemperatureCorrection,
    ContentType,
    DeepBlack,
    DscMode,
    DynamicControl,
    EcoMode,
    EShift,
    FilmmakerMode,
    GraphicMode,
    Hdmi2D3D,
    HdmiColorSpace,
    HdmiInputLevel,
    Hdr,
    HdrLevel,
    HdrProcessing,
    HideEco,
    HighlightColor,
    Input,
    InstallationMode,
    IntelligentLensAperture,
    LaserPower,
    LightPower,
    LightTime,
    LinkRate,
    LowLatencyAutoMode,
    LowLatencyMode,
    MacAddress,
    Mask,
    ModelName,
    MotionEnhance,
    OffTimer,
    PictureMode,
    PictureModeHdr10,
    PictureModeHlg,
    PictureModeSdr,
    PictureModeSdr3d,
    Power,
    Remote,
    Signal,
    Smoother,
    Source,
    TheaterOptimizer,
    TheaterOptimizerLevel,
    TheaterOptimizerProcessing,
    Trigger,
    Version,
)
